/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.2)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: l_imath.hpp,v 1.16 2006/12/06 13:11:25 cxsc Exp $ */

#ifndef LX_IMATH_HPP_L_INCLUDED
#define LX_IMATH_HPP_L_INCLUDED

#include "lx_interval.hpp"
#include <l_imath.hpp>

namespace cxsc {

// ---------------- lx_interval Konstanten --------------------------------- 

    //! Enclosure-Interval for \f$ \pi \f$
    lx_interval Pi_lx_interval() throw(); // Pi
    //! Enclosure-Interval for \f$ \ln(2) \f$
    lx_interval Ln2_lx_interval() throw();
    //! Enclosure-Interval for \f$ \ln(10) \f$
    lx_interval Ln10_lx_interval() throw();
    //! Enclosure-Interval for \f$ 1/\pi \f$
    lx_interval Pir_lx_interval() throw();
    //! Enclosure-Interval for \f$ \sqrt{\pi} \f$
    lx_interval SqrtPi_lx_interval() throw();
    //! Enclosure-Interval for \f$ \sqrt{2\pi} \f$
    lx_interval Sqrt2Pi_lx_interval() throw();
    //! Enclosure-Interval for \f$ \sqrt{2} \f$
    lx_interval Sqrt2_lx_interval() throw();
    //! Enclosure-Interval for \f$ \sqrt{3} \f$
    lx_interval Sqrt3_lx_interval() throw();
    //! Enclosure-Interval for \f$ 1/\ln(2) \f$
    lx_interval Ln2r_lx_interval() throw();
    //! Enclosure-Interval for \f$ \pi/3 \f$
    lx_interval Pid3_lx_interval() throw();
    //! Enclosure-Interval for \f$ 1/\sqrt{\pi} \f$
    lx_interval SqrtPir_lx_interval() throw();
    //! Enclosure-Interval for \f$ 1/\sqrt{2\pi} \f$
    lx_interval Sqrt2Pir_lx_interval() throw();
    //! Enclosure-Interval for \f$ \ln(\pi) \f$
    lx_interval LnPi_lx_interval() throw();
    //! Enclosure-Interval for \f$ \ln(2\pi) \f$
    lx_interval Ln2Pi_lx_interval() throw();
    //! Enclosure-Interval for \f$ e=2.718... \f$
    lx_interval E_lx_interval() throw();
    //! Enclosure-Interval for \f$ e^{\pi} \f$
    lx_interval EpPi_lx_interval() throw();
    //! Enclosure-Interval for \f$ \mbox{EulerGamma}=0.5772... \f$
    lx_interval EulerGamma_lx_interval() throw();
    //! Enclosure-Interval for \f$ \mbox{Catalan=0.9159... \f$
    lx_interval Catalan_lx_interval() throw();
    //! Enclosure-Interval for \f$ \sqrt{5} \f$
    lx_interval sqrt5_lx_interval() throw();
    //! Enclosure-Interval for \f$ \sqrt{7} \f$
    lx_interval sqrt7_lx_interval() throw();
    //! Enclosure-Interval for \f$ 1-2^{-2097} \f$
    lx_interval One_m_lx_interval() throw();
    //! Enclosure-Interval for \f$ 1+2^{-2097} \f$
    lx_interval One_p_lx_interval() throw();

// ---------------- lx_interval Funktionen ---------------------------------

//! Calculates \f$ \sqrt{[x]}  \f$
lx_interval sqrt(const lx_interval&) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ [x]^2  \f$
lx_interval sqr(const lx_interval&) throw();
//! Calculates \f$ \ln([x]) \f$
lx_interval ln(const lx_interval &) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \ln(1+[x]) \f$
lx_interval lnp1(const lx_interval &) throw();
//! Calculates \f$ \exp([x]) \f$
lx_interval exp(const lx_interval &) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ 2^{[x]} \f$
lx_interval exp2(const lx_interval &) throw(); // 2^x
//! Calculates \f$ 10^{[x]} \f$
lx_interval exp10(const lx_interval &) throw(); // 10^x
//! Calculates \f$ \exp([x])-1 \f$
lx_interval expm1(const lx_interval &x) throw(); 
//! Calculates \f$ [x]^n \f$
lx_interval power(const lx_interval &, int);
//! Calculates \f$ [x]^{[y]} \f$
lx_interval pow(const lx_interval &, const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ (1+[x])^{[y]} \f$
lx_interval xp1_pow_y(const lx_interval &, const lx_interval &) 
                        throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF); // (1+x)^y
//! Calculates \f$ \sin([x]) \f$
lx_interval sin(const lx_interval &);
//! Calculates \f$ \sin(n\cdot\pi+[x]) \f$
lx_interval sin_n(const lx_interval &x, int n);
//! Calculates \f$ \cos([x]) \f$
lx_interval cos(const lx_interval &);
//! Calculates \f$ \cos((n+1/2)\cdot\pi+[x]) \f$
lx_interval cos_n(const lx_interval &x, int n);
//! Calculates \f$ \tan([x]) \f$
lx_interval tan(const lx_interval &) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \cot([x]) \f$
lx_interval cot(const lx_interval &) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \sqrt{1+[x]^2} \f$
lx_interval sqrt1px2(const lx_interval &) throw();
//! Calculates \f$ \arctan([x]) \f$
lx_interval atan(const lx_interval &) throw();
//! Calculates \f$ \sqrt{1-[x]^2} \f$
lx_interval sqrt1mx2(const lx_interval &) throw();
//! Calculates \f$ \sqrt{[x]^2-1} \f$
lx_interval sqrtx2m1(const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \arcsin([x]) \f$
lx_interval asin(const lx_interval & ) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \arccos([x]) \f$
lx_interval acos(const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \mbox{arccot}([x]) \f$
lx_interval acot(const lx_interval &);
//! Calculates \f$ \sinh([x]) \f$
lx_interval sinh(const lx_interval &);
//! Calculates \f$ \cosh([x]) \f$
lx_interval cosh(const lx_interval &);
//! Calculates \f$ \tanh([x]) \f$
lx_interval tanh(const lx_interval &);
//! Calculates \f$ \coth([x]) \f$
lx_interval coth(const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \sqrt{[x]+1}-1} \f$
lx_interval sqrtp1m1(const lx_interval &) throw();
//! Calculates \f$ \mbox{arcsinh}([x]) \f$
lx_interval asinh(const lx_interval &) throw();
//! Calculates \f$ \mbox{arccosh}([x]) \f$
lx_interval acosh(const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \mbox{arccosh}(1+[x]) \f$
lx_interval acoshp1(const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \mbox{arctanh}([x]) \f$
lx_interval atanh(const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \mbox{arctanh}(1-[x]) \f$
lx_interval atanh1m(const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \mbox{arctanh}(-1+[x]) \f$
lx_interval atanhm1p(const lx_interval &) throw();
//! Calculates \f$ \mbox{arccoth}([x]) \f$
lx_interval acoth(const lx_interval &) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \mbox{arctanh}(+1+[x]) \f$
lx_interval acothp1(const lx_interval & ) 
                                   throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF);
//! Calculates \f$ \mbox{arctanh}(-1-[x]) \f$
lx_interval acothm1m(const lx_interval &) throw();
//! Calculates \f$ \sqrt{[x]^2 + [y]^2} \f$
lx_interval sqrtx2y2(const lx_interval &, const lx_interval &) throw();
//! Calculates \f$ \sqrt[n]{[x]} \f$
lx_interval sqrt(const lx_interval &, int) throw();
//! Calculates \f$ \ln(\sqrt{[x]^2 + [y]^2}) \f$
lx_interval ln_sqrtx2y2(const lx_interval &, const lx_interval &) throw();

} // namespace cxsc

#endif
