/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.2)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  F. Blomquist, University of Wuppertal, 19.09.2007;
*/

/* CVS $Id: lx_interval.cpp,v 1.4 2006/12/06 18:36:17 cxsc Exp $ */

#include "lx_interval.hpp"
#include "idot.hpp"

#include <iostream>

namespace cxsc {

void Dec2Bin(int p, const l_interval &x, int &ex, l_interval &li) throw()  
// For the given point interval x and for the given
// exponent p to base 10 the exponent ex to base 2 and the
// interval li are calculated to realize the inclusion
// 10^p * x subsetequal 2^ex * li; ( 10^p * x <= 2^ex * li );
// x must be a point interval. In general li will not be a point interval! 
// This function is for the internal use only.
// Blomquist, 18.01.2008;

{
    interval z;
    real r;
    double dbl;
    bool neg;
    int stagsave;

    ex = 0;  li = 0;
    stagsave = stagprec;
    z = x;
    r = Sup(abs(z));

    if (r != 0)
    {
	li = x;
	if (p!=0)
	{
	    neg = Inf(z)<0;
	    r = ( ln(r) + sub_int(p,307)*Ln10_real ) / Ln2_real;
	    if (abs(r)>2147483642)
	    {
		std::cerr << 
		    "ERROR in lx_interval(const int& n, const string& s); s out of range." 
			  << std::endl;
		exit(1); 
	    }
	    dbl = _double(r);
	    ex = (int) dbl;
	    stagprec = stagprec + 1;
	    if (neg) li = -li;
	    li = ln(li);
	    li = ( p*Ln10_l_interval() - ex*Ln2_l_interval() ) + li;
	    li = exp(li);
	    if (neg) li = -li;
	    stagprec = stagsave;
	    li = adjust(li);
	}
    }
} 

lx_interval::lx_interval(int n, const string &s) throw()
// Constructor:
// (10^n, string) --- > (ex,li) vom Typ lx_interval
// string s must be of the form:  
//        s = "[1.234567,1.234567]";   or:  s = "[0,0]";
// or:    s = "[1.234567e+20,1.234568e+20]";
// or:    s = "[1.234567e-20,1.234568e-20]";
// or:    s = "[0,1]";   or: s = "[-1,0]";  or: s = "[-1,+1]";
// (ex,li) is an optimal inclusion of  10^n*s;

// Not allowed strings:
//        s = "1.234567";     or:   s = "1.234567e+20";
// or:    s = "[1.2,1.2";     or:   s = "1.2,1.2]";
// or:    s = "+[1.2,1.2]";   or:   s = "-[1.2,1.2]";

{
    l_interval x(0),xh,a1,a2;
    lx_interval u; 
    int p,q; 

    // Check for brackets [...] in string s:
    string str1 = "[", str2 = "]";
    q = s.find_first_of(str1);
    p = s.find_first_of(str2);
    if (q == -1 || p == -1)
    {
	std::cerr << 
	    "ERROR in lx_interval(const int& n, const string& s); missing [...] in s." 
		  << std::endl;
	exit(1); 
    }

    x = x + 0;  // To get StagPrec(x) = actual value of stagprec;
    s >> x;
    xh = l_interval(Inf(x));
    Dec2Bin(n,xh,p,a1);

    xh = l_interval(Sup(x));
    Dec2Bin(n,xh,q,a2);

    u = lx_interval(lx_real(p,Inf(a1)),lx_real(q,Sup(a2)));

    ex = expo(u);
    li = li_part(u);
}


void scale_down(lx_interval &a)
// scale_down skaliert a moeglichst weit nach unten, ohne
// dass dabei binaere Stellen verloren gehen.
// Der mathematische Wert von a bleibt dabei erhalten, d.h.
// 2^(a.ex)*a.li  hat vor und nach der Skalierung stets den
// gleichen mathematischen Wert.
{
    int d;
    d = -1021 - expo_sm(a.li);
    // In case of a.li=0 we get d>0, and so nothing is done! 
    if (d<0)
    {
	Times2pown(a.li,d);
	a.ex = sub_int(a.ex,d);
    }
}

void scale_up(lx_interval &a)
// scale_up skaliert a.li moeglichst weit nach oben.  
// Nach der Skalierung wird jedoch  a.li + a.li  keinen Overflow 
// erzeugen.
// 2^(a.ex)*a.li  hat fuer d>0 vor und nach der Skalierung stets
// den gleichen Wert. Nur im Fall d<0 kann es durch die 
// Skalierung nach unten nur zu einer minimalen Aufblaehung kommen,
// denn im Fall  d<0  nimmt d nur die Werte d = -1,-2 an.   
{
    int d;
    d = 1022 - expo_gr(a.li);
    if (d>-4) // if (d>-3) ist auch moeglich!
    {
	Times2pown(a.li,d); // a.li wird mit 2^d multipliziert.
	a.ex = sub_int(a.ex,d); // damit 2^(a.ex)*a.li konstant bleibt.
    }
}


l_interval & l_interval::operator = (const lx_interval &a) throw()
{
    int p(expo(a)),exa;
    l_interval al(li_part(a));
    exa = expo_gr(al);
    if (exa<-100000) return (*this) = al; // al = 0;
    exa = add_int(p,exa);
    if (exa > 1024)
    {
	std::cerr << "ERROR in = operator:  Value of type lx_interval is unrealizable as l_interval." 
		  << std::endl;
        exit(1); 
    }

    Times2pown(al,p);
    return (*this) = al;
}

interval & interval::operator = (const lx_interval &a) throw()
{
    l_interval al;
    interval z;
    al = a;
    z = al;
    return (*this) = z;
}

lx_interval expo2zero(const lx_interval &a) throw()
// Der Zweierexponent von a wird auf Null gesetzt.
// Der Rueckgabewert res ist eine Einschliessung von a,
// falls kein Overflow entsteht.
{
    lx_interval res(0);
    l_interval la(li_part(a));
    int na(expo(a)), exa(expo_gr(la));

    if (exa > -100000) // a != 0;
	{
	    exa = add_int(exa,na);
	    if (exa > 1024)
	    {
		std::cerr << "ERROR in: lx_interval expo2zero(const lx_interval &a): Overflow!" << std::endl;
		exit(1);
	    }
	    Times2pown(la,na);
	    res = lx_interval(0,la);
	}

    return res;
}

lx_interval::lx_interval(const lx_real& a, const lx_real& b) throw()
{
    lx_real a_(a),b_(b);
    l_real lr;
    interval z;
    bool a_zero(eq_zero(a)), b_zero(eq_zero(b));
    int na,nb,d;

    if (a_zero || b_zero)
	if (a_zero) 
	{
	    ex = expo(b);
	    li = l_interval(0,lr_part(b));
	} else
	{
	    ex = expo(a);
	    li = l_interval(lr_part(a),0);
	}
    else
    {  // Now we have:  a,b != 0:
	scale_up(a_);  scale_up(b_);
	na = expo(a_);  nb = expo(b_);

	if (na >= nb)
	{
	    if (na==nb) // haeufigster Fall:
	    {
		ex = na;
		li = l_interval(lr_part(a_),lr_part(b_));
	    }
	    else // na > nb:
	    {
		d = sub_int(na,nb); // d = na-nb>0
		ex = na;
		lr = lr_part(b_);
		Times2pown(lr,z,-d);
                // Calculation of an upper bound of lr+z:
		if (z!=0) 
		    lr = Sup(l_interval(lr) + Sup(z));
		li = l_interval(lr_part(a_),lr);
	    }
	}
	else // na < nb:
	{
	    d = sub_int(nb,na); // d = nb-na>0
	    ex = nb;
	    lr = lr_part(a_);
	    Times2pown(lr,z,-d);
	    // Calculation of an lower bound of lr+z:
		if (z!=0)
		    lr = Inf(l_interval(lr) + Inf(z));
	    li = l_interval(lr,lr_part(b_));
	}
    }
}

lx_interval operator + (const lx_interval &a, const lx_interval &b) throw()
{
    int stagsave = stagprec,
	stagmax = 39,
	exa,exb,sa,sb,p,n;
    if (stagprec>stagmax) stagprec = stagmax;

    l_interval a1,b1;
    lx_interval res;
    interval z;

    a1 = li_part(a);  b1 = li_part(b);
    exa = expo_gr(a1);
    exb = expo_gr(b1);

    if (exa<-100000) return b;  // Ein Summand ist Null
    if (exb<-100000) return a;  // Ein Summand ist Null

    // a,b <> 0:
    sa = add_int(exa,expo(a));
    sb = add_int(exb,expo(b));
    if (sa > sb) // |a| >= |b|
    {
	p = 1022 - exa; // p >= 0
	Times2pown(a1,p);
	n = sub_int(expo(a),p);
	if (n>0 && expo(b)<sub_int(n,10000))
	{
	    z = b1;
	    if (Sup(z)<=0.0)
		b1 = l_interval(-MinReal,0);
	    else
		if (Inf(z)>=0.0)
		b1 = l_interval(0,MinReal);
		else
		    b1 = l_interval(-MinReal,MinReal);
	}
	else
	{
	    p = sub_int(expo(b),n); 
	    Times2pown(b1,p);
	}
    } else
    {
	p = 1022 - exb; // p >= 0
	Times2pown(b1,p);
	n = sub_int(expo(b),p);
	if (n>0 && expo(a)<sub_int(n,10000))
	{
	    z = a1;
	    if (Sup(z)<=0.0)
		a1 = l_interval(-MinReal,0);
	    else
		if (Inf(z)>=0.0)
		a1 = l_interval(0,MinReal);
		else
		    a1 = l_interval(-MinReal,MinReal);
	}
	else
	{
	    p = sub_int(expo(a),n);
	    Times2pown(a1,p);
	}
    }
    a1 = a1 + b1; // norm. staggered Addition
    res = lx_interval(n,a1);

    stagprec = stagsave;
    res = adjust(res);
    return res;
}

lx_interval operator * (const lx_interval &a, const lx_interval &b) throw()
{
    int stagsave = stagprec,
	stagmax = 39,
	exa,exb,na,nb,d,D,k;

    if (stagprec > stagmax) stagprec = stagmax;

    l_interval al,bl;
    lx_interval a_(a),b_(b),res(0);

    scale_down(a_);  scale_down(b_);

    al = li_part(a_);  bl = li_part(b_);
    exa = expo_gr(al);
    exb = expo_gr(bl);

    if (exa>-100000 && exb>-100000) // Beide Faktoren sind <> 0:
    {
	if (exa+exb <= 1022)
	{
	    if (exa<0)
	    {
		Times2pown(al,-exa);  
		na = add_int(expo(a_),exa);  // a = 2^na * al;
		d = 1022 - exb;
	    }
	    else // exa >= 0
	    {
		na = expo(a_);
		d = 1022 - (exa+exb);
	    }
	    Times2pown(bl,d);
	    nb = sub_int(expo(b_),d);  // b = 2^nb * bl;
	}
	else // exa+exb > 1022
	{
	    d = exa + exb - 1022;  // d > 0;
	    if (exa>exb)
	    {
		D = exa - exb; // D > 0;
		if (d <= D)
		{
		    Times2pown(al,-d);
		    na = add_int(expo(a_),d); // a = 2^na * al;
		    nb = expo(b_);
		}
		else // d > D;
		{
		    k = d - D;
		    if (k%2 != 0) k++;
		    k = k/2;
		    Times2pown(bl,-k);
		    nb = add_int(expo(b_),k); // b = 2^nb * bl;
		    k = k + D;
		    Times2pown(al,-k);
		    na = add_int(expo(a_),k); // a = 2^na * al;
		}
	    }
	    else // exb >= exa
	    {
		D = exb - exa; // D > 0;
		if (d <= D)
		{
		    Times2pown(bl,-d);
		    nb = add_int(expo(b_),d); // b = 2^nb * bl;
		    na = expo(a_);
		}
		else // d > D
		{
		    k = d - D;
		    if (k%2 != 0) k++;
		    k = k/2;
		    Times2pown(al,-k);
		    na = add_int(expo(a_),k);  // a = 2^na * al;
		    k = k + D;
		    Times2pown(bl,-k);
		    nb = add_int(expo(b_),k);  // b = 2^nb * bl;
		}
	    }
	}
	al = al*bl;      // norm. staggered Multiplikation
	na = add_int(na,nb);
	res = lx_interval(na,al);
    }
    stagprec = stagsave;
    res = adjust(res);
    return res;
} // operator *

lx_interval operator / (const lx_interval &a, const lx_interval &b) 
                              throw(ERROR_LINTERVAL_DIV_BY_ZERO)
{
    int stagsave = stagprec,
	stagmax = 39,
	exa,exb,na,nb,d;
    if (stagprec>stagmax) stagprec = stagmax;

    l_interval al,bl;
    lx_interval a_(a),b_(b),res;

    scale_down(a_);  scale_down(b_);
    al = li_part(a_);  bl = li_part(b_);

    exa = expo_gr(al);
    exb = expo_gr(bl);

    if (exb < -100000) 
	cxscthrow(ERROR_LINTERVAL_DIV_BY_ZERO("lx_interval operator/(const lx_interval &a, const lx_interval &b)"));

    if (exa < -100000) return a;
    // Now it holds:  a,b <> 0:

    bool bl_point( point_intv(bl) );

    // First: scaling the nominator al to 1022:
    d = 1022 - exa;
    Times2pown(al,d);   // expo_gr(al) = 1022;
    na = sub_int(expo(a_),d);  // a = 2^na * al;

    if (bl_point) 
    {
	if (0<=exb && exb<=511) nb = expo(b_);
	else 
	    if (exb<0)
	    {  // scaling the denominator bl to 0:
		Times2pown(bl,-exb);
		nb = add_int(expo(b_),exb);  // b = 2^nb * bl;
	    }
	    else // exb > 511;
                 // Now scaling the denominator bl to 511:
	    {
		d = 511 - exb;
		Times2pown(bl,d);
		nb = sub_int(expo(b_),d);
	    }
    }
    else // scaling the denominator bl to 511:
    {
	d = 511 - exb;
	Times2pown(bl,d);
	nb = sub_int(expo(b_),d);
    }

    al = al / bl;      // norm. staggered division;
    na = sub_int(na,nb);
    res = lx_interval(na,al);

    stagprec = stagsave;
    res = adjust(res);
    return res;
}

// ----------------------- Input --------------------------------------------

/*
// Funktionierende Version mit Zweier-Exponenten:
std::string & operator >> (std::string &s, lx_interval &a) 
    throw(ERROR_CINTERVAL_EMPTY_INTERVAL)
// Writs strings s to variable a of type lx_interval 
// and returns an empty string s;
{
    l_interval la;
    real exr;
    double dbl;
    int exa;

    s = skipwhitespacessinglechar (s, '{');
    s = skipwhitespacessinglechar (s, '(');
    s >> SaveOpt >> exr; // writing int-value to real-value
    dbl = _double(exr);  // exr to double
    exa = (int) dbl;     // double to int
    s = skipwhitespacessinglechar (s, ')');
    s = skipwhitespacessinglechar (s, ',');
    s = s >> la >> RestoreOpt;
    s = skipwhitespacessinglechar (s, '}');
    s = skipwhitespaces (s); // erasing whitespaces
    a = lx_interval(exa,la);
    if (Inf(la)>Sup(la)) 
	cxscthrow(ERROR_CINTERVAL_EMPTY_INTERVAL(
		  "string & operator >> (std::string &s, lx_interval &a)"));
    return s;
}
*/

// Neue Version mit Zehner-Exponenten
std::string & operator >> (std::string &s, lx_interval &a) 
    throw(ERROR_CINTERVAL_EMPTY_INTERVAL)
// Writs strings s to variable a of type lx_interval 
// and returns an empty string s;
{
    l_interval la;
    real exr;
    double dbl;
    int i,exa;

    s = skipwhitespacessinglechar (s, '{');
//    s = skipwhitespacessinglechar (s, '(');
    s >> exr; // writing int-value to real-value
    dbl = _double(exr);  // exr to double
    exa = (int) dbl;     // double to int
//    s = skipwhitespacessinglechar (s, ')');
    s = skipwhitespacessinglechar (s, ',');
    i = s.find("]");
    s.erase(i+1); 

    a = lx_interval(exa,s);
    s = "";
    la = li_part(a);
 
    if (Inf(la)>Sup(la)) 
	cxscthrow(ERROR_CINTERVAL_EMPTY_INTERVAL(
		  "string & operator >> (std::string &s, lx_interval &a)"));
    return s;
}

void operator >> (const std::string &s, lx_interval &a) throw()
{
// Writs strings s to variable a of type lx_interval;
    std::string r(s);
    r >> a;
}

void operator >> (const char *s, lx_interval &a) throw()
{
    std::string r(s);
    r >> a;
}

/*
// Funktionierende Version fuer Zweier-Exponenten:
std::istream & operator >> (std::istream &s, lx_interval &a) 
	throw(ERROR_CINTERVAL_EMPTY_INTERVAL)
{
    l_interval la;
    real exr;
    double dbl;
    int exa;
    char c;

    skipeolnflag = inpdotflag = true;
    c = skipwhitespacessinglechar (s, '{');
    if (inpdotflag) s.putback(c);
    c = skipwhitespacessinglechar (s, '(');
    if (inpdotflag) s.putback(c);
    s >> SaveOpt >> exr; // writing int-value to real-value
    dbl = _double(exr);  // exr to double
    exa = (int) dbl;     // double to int
    c = skipwhitespacessinglechar (s, ')');
    if (inpdotflag) s.putback(c);

    skipeolnflag = inpdotflag = true;
    c = skipwhitespacessinglechar (s, ',');
    if (inpdotflag) s.putback(c);
    skipeolnflag = inpdotflag = true;
    s >> la >> RestoreOpt;

    a = lx_interval(exa,la);

    if (!waseolnflag)
    {
	skipeolnflag = false; inpdotflag = true;
	c = skipwhitespaces (s);
	if (inpdotflag && c != '}')
	    s.putback(c);
    }

    if (Inf(la)>Sup(la)) 
	cxscthrow(ERROR_CINTERVAL_EMPTY_INTERVAL(
		  "istream & operator >> (std::istream &s, lx_interval &a)"));
    return s;
}
*/

bool StrContains(const string &s, const char &a, const char &b)
// returns 1, if s contains char a or char b, otherwise 0;
{
    int ia,ib;
    ia = s.find(a,0);
    ib = s.find(b,0);
    return (ia>=0 || ib>=0);
}

// Funktionierende Version fuer Zehner-Exponenten:
std::istream & operator >> (std::istream &s, lx_interval &a) 
	throw(ERROR_CINTERVAL_EMPTY_INTERVAL)
{
    char c;
    int exa,exr;
    double dbl;
    string str,sh;

    skipeolnflag = inpdotflag = true;
    c = skipwhitespacessinglechar (s, '{');
    if (inpdotflag) s.putback(c);
    s >> SaveOpt >> exr; // writing int-value to real-value
    dbl = _double(exr);  // exr to double
    exa = (int) dbl;     // double to int
    c = skipwhitespacessinglechar (s, ',');
    if (inpdotflag) s.putback(c);
    skipeolnflag = inpdotflag = true;
    s >> str >> RestoreOpt;

    while ( !StrContains(str,']','}') )
    {
	c = skipwhitespaces (s);
	if (inpdotflag && c != '}') s.putback(c);
	if (c == '}') break;
	if (c == ' ') break;
	s >> sh;
	str = str + sh;
    }

    a = lx_interval(exa,str);

    if (!waseolnflag)
    {
	skipeolnflag = false; inpdotflag = true;
	c = skipwhitespaces (s);
	if (inpdotflag && c != '}')
	    s.putback(c);
    }

    return s;
}


// ----------------------- Output --------------------------------------------

void lx_intv_point2Dec(const lx_interval &a, int &p, l_interval &m)
// For a given point interval a the two values p and m are
// calculated to realize:
// a <= 10^p * m;
// Blomquist, 17.01.08; 
{
    int aex(expo(a)), stagsave;
    real r;
    double dbl;
    interval z;
    bool neg;

    m = 0.0;  p = 0;
    z = li_part(a);
    if ( !(0<=z) )
    {
	r = Sup(abs(z));
	
	neg = (Inf(z)<0);
	if (r<1e-304)
	{
	    times2pown(r,67);
	    r = ln(r) - 67*Ln2_real;
	    r = (aex*Ln2_real + r) / Ln10_real;
	}
	else
	    r = (aex*Ln2_real + ln(r)) / Ln10_real;

	dbl = _double(r);
	p = (int) dbl;
	p--;  // This is the new exponent to base 10;

	// Now calculatiing the interval m:
	stagsave = stagprec;
	stagprec = stagprec+1;

	m = li_part(a);
	if (neg) m = -m;

	if (Sup(m)<1e-304)
	{
	    times2pown(m,67);
	    m = ln(m) - 67*Ln2_l_interval();
	}
	else
	{   
	    m = ln(m); // Vorher!
//	        m = lnp1(m-1);
	}
	m = (aex*Ln2_l_interval()-p*Ln10_l_interval()) + m;
	m = exp(m);
	if (neg) m = -m;

	stagprec = stagsave;
	m = adjust(m);
    }
} // lx_intv_point2Dec(...)

void lx_intv2Dec(const lx_interval &a, int &p, l_interval &m)
// For a given wide interval a the two values p and m are
// calculated to realize the inclusion:
// a <= 10^p * m;
// Blomquist, 17.01.08;
{
    int p1,p2,d,stagsave;
    l_interval m1,m2;
    lx_interval x;
    string s;

    x = lx_interval(Inf(a));
    lx_intv_point2Dec(x,p1,m1);

    x = lx_interval(Sup(a));
    lx_intv_point2Dec(x,p2,m2);

    stagsave = stagprec;
    stagprec = stagprec + 1;

    if (m1==0.0)
    {
	p = p2;
	m = l_interval(0,Sup(m2));
    }
    else
	if (m2==0.0)
	{
	    p = p1;
	    m = l_interval(Inf(m1),0);
	}
	else
	{  // m1 != 0  and  m2 != 0
	    p = (p1>p2)? p1 : p2; // p = max(p1,p2);
	    s = "[1,1]";
	    if (p==p1)
	    {
		d = p2-p1;
		if (d!=0)
		{
		    x = lx_interval(d,s); // x = 10^d
		    x = x * m2;
		    m2 = x;  // m2 = (old m2) * 10^d;
		}
	    } else // p = p2
	    {
		d = p1-p2;
		if (d!=0)
		{
		    x = lx_interval(d,s);
		    x = x * m1;
		    m1 = x;  // m1 = (old m1) * 10^d;
		}
	    }
	    m = l_interval(Inf(m1),Sup(m2));
	}
    stagprec = stagsave;
    m = adjust(m);
} // lx_intv2Dec(...)

std::ostream& operator << (std::ostream& s,const lx_interval& a) throw()
// An interval a of type lx_interval is written to the 
// output channel in the decimal form: 
// { exponent p to base 10, l_interval m } = 10^p * m;
// The following inclusion is realized:   a <= 10^p * m;
// Blomquist, 18.01.2008;
{   
    l_interval m;
    int p;

    lx_intv2Dec(a, p, m);

    s << '{'          
    << p << ", "   
    << m       
    << '}';

    return s;
}

std::string & operator << (std::string &s, const lx_interval &a) throw()
// The value of a variable a of type lx_interval is copied to a string s.
// s has the form:  {(ex), li} = 2^ex * li;
{  
    std::stringstream ss;
    string str;
    s+="{(";
    ss << a.ex;
    ss >> str;
    s += str;
    s+="), ";
    s << a.li; 
    s+='}';
    return s;
}


// ----- Help Functions, declared in lx_interval.hpp outside ---------------
// --------------------- the class lx_interval------------------------------

l_interval point_max(void)
// returns a staggered point interval with maximum exponent 1020,
// whereby nearly all mantissa bits are set. 
{
    l_real lr;
    l_interval li = sqrt(l_interval(3.995)), res(li);
    times2pown(li,1019);
    lr = Inf(li);
    lr = ((lr + Inf(res)) + MinReal) + minreal;
    res = l_interval(lr);
    return res;
}

l_interval point_any(int n)
// returns a staggered point interval with exponent n,
// whereby nearly all mantissa bits are set.
// -1074 <= n <= +1020; 
{
    int n_(n);
    l_interval res;

    if (n_>1020)  n_ = 1020;
    if (n_<-1074) n_ = -1074;
    res = point_max();
    times2pown(res,n_ - 1020);
    return l_interval(Inf(res+MinReal));
}

l_interval wide_max(void)
// returns a staggered interval a with maximum exponent 1020
// and diam(a)>0, whereby nearly all mantissa bits are set. 
{
    l_interval a = point_max();
    return a + interval(MinReal,2*MinReal);
}

l_interval wide_any(int n)
// returns a staggered point interval a with exponent n,
// and diam(a)>0, whereby nearly all mantissa bits are set.
// -1074 <= n <= +1020;
{
    l_interval a = point_any(n);
    return a + interval(MinReal,2*MinReal);
}

/* ---------------- Array of constants ------------------ */

// ln_N[0] = ln(2); ln_N[1] = ln(3); ... ln_N[179] = ln(181);

real ln_N[180] = 
{ ln(2), ln(3), ln(4), ln(5), ln(6), ln(7), ln(8), ln(9), ln(10),
  ln(11),ln(12),ln(13),ln(14),ln(15),ln(16),ln(17),ln(18),ln(19),ln(20),
  ln(21),ln(22),ln(23),ln(24),ln(25),ln(26),ln(27),ln(28),ln(29),ln(30),
  ln(31),ln(32),ln(33),ln(34),ln(35),ln(36),ln(37),ln(38),ln(39),ln(40),
  ln(41),ln(42),ln(43),ln(44),ln(45),ln(46),ln(47),ln(48),ln(49),ln(50),
  ln(51),ln(52),ln(53),ln(54),ln(55),ln(56),ln(57),ln(58),ln(59),ln(60),
  ln(61),ln(62),ln(63),ln(64),ln(65),ln(66),ln(67),ln(68),ln(69),ln(70),
  ln(71),ln(72),ln(73),ln(74),ln(75),ln(76),ln(77),ln(78),ln(79),ln(80), 
  ln(81),ln(82),ln(83),ln(84),ln(85),ln(86),ln(87),ln(88),ln(89),ln(90),
  ln(91),ln(92),ln(93),ln(94),ln(95),ln(96),ln(97),ln(98),ln(99),ln(100),
  ln(101),ln(102),ln(103),ln(104),ln(105),ln(106),ln(107),ln(108),ln(109),
  ln(110),ln(111),ln(112),ln(113),ln(114),ln(115),ln(116),ln(117),ln(118),
  ln(119),ln(120),ln(121),ln(122),ln(123),ln(124),ln(125),ln(126),ln(127),
  ln(128),ln(129),ln(130),ln(131),ln(132),ln(133),ln(134),ln(135),ln(136),
  ln(137),ln(138),ln(139),ln(140),ln(141),ln(142),ln(143),ln(144),ln(145),
  ln(146),ln(147),ln(148),ln(149),ln(150),ln(151),ln(152),ln(153),ln(154),
  ln(155),ln(156),ln(157),ln(158),ln(159),ln(160),ln(161),ln(162),ln(163),
  ln(164),ln(165),ln(166),ln(167),ln(168),ln(169),ln(170),ln(171),ln(172),
  ln(173),ln(174),ln(175),ln(176),ln(177),ln(178),ln(179),ln(180),ln(181)};
// ----------------------------------------------------------------

} // namespace cxsc
