/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.2)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  F. Blomquist, University of Wuppertal, 19.09.2007;
*/
/* CVS $Id: lx_interval.hpp,v 1.2 2006/10/05 10:09:53 cxsc Exp $ */

#ifndef LX_INTERVAL_HP_INCLUDED
#define LX_INTERVAL_HP_INCLUDED

#include <l_interval.hpp>
#include <l_imath.hpp>
#include "lx_real.hpp"

namespace cxsc {

class lx_interval
{
private:
    // ------------- Data Elements -------------------------------------------
    int ex;
    l_interval li;
    // The mathematical value of an object of type lx_interval is
    // interpreted as:  2^(ex) * li;

public:
    // ------------- Constructors --------------------------------------------

    //! Constructor of class lx_interval
    lx_interval(void)  throw() {}
    //! Constructor of class lx_interval
    lx_interval(int n, const l_interval& a) throw() : ex(n), li(a) { }
    //! Constructor of class lx_interval
    lx_interval(int n, const l_real& a)     throw() : ex(n), li(a) { }
    //! Constructor of class lx_interval
    lx_interval(int n, const interval& a)   throw() : ex(n), li(a) { }
    //! Constructor of class lx_interval
    explicit lx_interval(int n, const real& a)throw() : ex(n), li(a) { }
    //! Constructor of class lx_interval
    explicit lx_interval(const l_interval& a) throw() : ex(0), li(a) { }
    //! Constructor of class lx_interval
    explicit lx_interval(const l_real& a)     throw() : ex(0), li(a) { }
    //! Constructor of class lx_interval
    lx_interval(const l_real& a, const l_real& b)  throw() : ex(0),li(a,b) { }
    //! Constructor of class lx_interval
    explicit lx_interval(const interval& a)   throw() : ex(0), li(a)   { }
    //! Constructor of class lx_interval
    explicit lx_interval(const real& a)       throw() : ex(0), li(a)   { }
    //! Constructor of class lx_interval
    explicit lx_interval(const real& a, const real& b) 
                                             throw() : ex(0), li(a,b) { }
    //! Constructor of class lx_interval
    lx_interval(const lx_real&, const lx_real&) throw();
    //! Constructor of class lx_interval
    explicit lx_interval(const lx_real& a)     
                                 throw() : ex(expo(a)), li(lr_part(a)) { }
    //! Constructor of class lx_interval
    lx_interval(int, const string&) throw(); 

    // ------------- Assignments ---------------------------------------------

    //! Implementation of standard assigning operator
    inline lx_interval & operator = (const lx_interval & a) throw();
    //! Implementation of standard assigning operator
    inline lx_interval & operator = (const l_interval & a) throw();
    //! Implementation of standard assigning operator
    inline lx_interval & operator = (const l_real & a)     throw();
    //! Implementation of standard assigning operator
    inline lx_interval & operator = (const real & a)       throw();
    //! Implementation of standard assigning operator
    inline lx_interval & operator = (const interval & a)   throw();
    //! Implementation of standard assigning operator
    inline lx_interval & operator = (const lx_real & a)     throw();

    // l_interval & operator = (const lx_interval &a) throw(); declared
    // in l_interval.hpp and implemented in lx_interval.cpp; 

    // ------------- Functions -----------------------------------------------

    //! Returns the actual precision of an interval of type lx_interval
    friend inline int StagPrec(const lx_interval&) throw(); 
    //! Returns the exponent of base 2 of an interval of type lx_interval
    friend inline int expo(const lx_interval&) throw();
    //! Returns the l_interval part of an interval of type lx_interval
    friend inline l_interval li_part(const lx_interval&) throw();

    //! The new interval a includs the old one and the old a.li is scaled downwards
    friend void scale_down(lx_interval &a);
    //! The new interval a includs the old one and the old a.li is scaled upwards
    friend void scale_up  (lx_interval &a);

    //! matches the precision of an interval to the actual stagprec value
    friend inline lx_interval adjust(const lx_interval &) throw();
    //! Returns the absolute value of an interval
    friend inline lx_interval abs(const lx_interval &) throw();
    //! Returns 1 if the interval is a point interval
    friend inline bool point_intv(const lx_interval &);
    //! Implementation of standard negation operation
    friend inline bool operator ! (const lx_interval &) throw(); 
    //! Returns 1 if an empty interval is present
    friend inline bool IsEmpty(const lx_interval &) throw(); 
    //! Multiplication of an interval with \f$ 2^n \f$
    friend inline void times2pown(lx_interval &, int) throw();
    //! Multiplication of an interval with \f$ 2^n, n<0, \f$ without integer overflow
    friend inline void times2pown_neg(lx_interval &, int);
    //! Returns the relative diameter of an interval
    friend inline l_real RelDiam( const lx_interval &);
    //! Returns the diameter of an interval
    friend inline lx_real diam(const lx_interval &) throw();
    //! Returns the rounded middle of an interval
    friend inline lx_real mid(const lx_interval &) throw();

    //! Returns the infimum of an interval
    friend inline lx_real Inf(const lx_interval &) throw();
    //! Returns the supremum of an interval
    friend inline lx_real Sup(const lx_interval &) throw();

    // ---------------- Monadic arithmetic operator ---------------------

    //! Implementation of standard algebraic negative sign operation
    friend inline lx_interval operator-(const lx_interval & a) throw();

    // ----------------------- Output -----------------------------------

    //! Implementation of standard output method
    friend std::ostream& operator << (std::ostream&, const lx_interval&) 
	throw();
    //! Implementation of standard output method
    friend std::string & operator << (std::string&, const lx_interval&) 
	throw();

}; // end of class lx_interval

// **********************************************************************
// **********************************************************************

// -------------------------------------------------------------------
//    Declaration of friend funktions outside the class lx_interval 
// -------------------------------------------------------------------

inline int StagPrec(const lx_interval &a) throw();
inline int expo(const lx_interval &a) throw();
inline l_interval li_part(const lx_interval &a) throw();
       void scale_down(lx_interval &);
       void scale_up  (lx_interval &);
inline bool point_intv(const lx_interval &);
inline bool IsEmpty(const lx_interval &) throw();
inline void times2pown(lx_interval &, int) throw();
inline void times2pown_neg(lx_interval &a, int n);
inline l_real RelDiam( const lx_interval & );
inline lx_real Inf(const lx_interval &) throw();
inline lx_real Sup(const lx_interval &) throw();
inline lx_interval abs(const lx_interval &) throw();
inline lx_interval adjust(const lx_interval &) throw();
inline lx_real diam(const lx_interval &) throw();
inline lx_real mid(const lx_interval &) throw();


// ------------------------ Input --------------------------------------

    //! Implementation of standard input method
    std::string & operator >> (std::string &s, lx_interval &a) 
	throw(ERROR_CINTERVAL_EMPTY_INTERVAL);
    //! Implementation of standard input method
    void operator >> (const std::string &s, lx_interval &a) throw();
    //! Implementation of standard input method
    void operator >> (const char *s, lx_interval&) throw();
    //! Implementation of standard input method
    std::istream & operator >> (std::istream&, lx_interval&) 
	throw(ERROR_CINTERVAL_EMPTY_INTERVAL);

// ------------------------ Output --------------------------------------

std::ostream& operator << (std::ostream& s,const lx_interval& a) throw();
// A value a of type lx_interval is written to the 
// output channel as decimal number in the form: 
// { exponent p to base 10, interval mantissa m } = 10^p * m;

std::string & operator << (std::string &s,const lx_interval& a) throw();
// The value of a variable a of type lx_interval is copied to a string s.
// s has the form:  {(ex), li} = 2^ex * li;


// -------------------------------------------------------------------
// ------- Function declarations outside the class lx_interval --------
// -------------------------------------------------------------------

//! b = expo2zero(a) returns \f$ a\subseteq b \f$ with  \f$ \verb+b.ex+=0 \f$
lx_interval expo2zero(const lx_interval &) throw();
//! Checks arguments for disjointness
inline int Disjoint(const lx_interval &, const lx_interval &);
//! Checks if first argument lies in the interior of second argument
inline int in (const lx_interval&, const lx_interval&);
//! Checks if first argument lies in the interior of second argument
inline int in (const l_interval&, const lx_interval&);
//! Checks if first argument lies in the interior of second argument
inline int in (const l_interval&, const lx_interval&);
//! Checks if first argument is part of second argument
inline int in (const interval&, const lx_interval&);

//! Checks if first argument is part of second argument       
inline int in (const lx_real&, const lx_interval&);
//! Checks if first argument is part of second argument       
inline int in (const l_real&, const lx_interval&);
//! Checks if first argument is part of second argument       
inline int in (const real&, const lx_interval&);

//! Returns an epsilon inflation of the first argument
inline lx_interval Blow( const lx_interval&, const real& );

//! Computes the smallest absolute value \f$ \left< \left[x\right] \right> \f$
inline lx_real AbsMin (const lx_interval&); 
//! Computes the greatest absolute value \f$ \left|\left[x\right]\right| \f$  
inline lx_real AbsMax (const lx_interval&); 

// -----------------------------------------------------------------------
// ------------- set comparisons -----------------------------------------
// -----------------------------------------------------------------------

// ---- lx_interval--lx_interval

//! Implementation of standard less-than operation
inline bool operator <  (const lx_interval&, const lx_interval&) throw();
//! Implementation of standard less-or-equal-than operation
inline bool operator <= (const lx_interval&, const lx_interval&) throw();
//! Implementation of standard greater-than operation
inline bool operator >  (const lx_interval&, const lx_interval&) throw();
//! Implementation of standard greater-or-equal-than operation
inline bool operator >= (const lx_interval&, const lx_interval&) throw();

// ---- lx_interval--l_interval

//! Implementation of standard less-than operation
inline bool operator <  (const lx_interval&, const l_interval&) throw();
//! Implementation of standard less-or-equal-than operation
inline bool operator <= (const lx_interval&, const l_interval&) throw();
//! Implementation of standard less-than operation
inline bool operator <  (const l_interval&, const lx_interval&) throw();
//! Implementation of standard less-or-equal-than operation
inline bool operator <= (const l_interval&, const lx_interval&) throw();
//! Implementation of standard greater-than operation
inline bool operator >  (const lx_interval&, const l_interval&) throw();
inline bool operator >= (const lx_interval&, const l_interval&) throw();
//! Implementation of standard greater-than operation
inline bool operator >  (const l_interval&, const lx_interval&) throw();
//! Implementation of standard greater-or-equal-than operation
inline bool operator >= (const l_interval&, const lx_interval&) throw();

// ---- lx_interval--interval

//! Implementation of standard less-than operation
inline bool operator <  (const lx_interval&, const interval&) throw();
//! Implementation of standard less-or-equal-than operation
inline bool operator <= (const lx_interval&, const interval&) throw();
//! Implementation of standard less-than operation
inline bool operator <  (const interval&, const lx_interval&) throw();
//! Implementation of standard less-or-equal-than operation
inline bool operator <= (const interval&, const lx_interval&) throw();
//! Implementation of standard greater-than operation
inline bool operator >  (const lx_interval&, const interval&) throw();
//! Implementation of standard greater-or-equal-than operation
inline bool operator >= (const lx_interval&, const interval&) throw();
//! Implementation of standard greater-than operation
inline bool operator >  (const interval&, const lx_interval&) throw();
//! Implementation of standard greater-or-equal-than operation
inline bool operator >= (const interval&, const lx_interval&) throw();

// ---- lx_interval--real

//! Implementation of standard less-than operation
inline bool operator <  (const real &, const lx_interval &) throw();
//! Implementation of standard less-or-equal-than operation
inline bool operator <= (const real &, const lx_interval &) throw();
//! Implementation of standard greater-than operation
inline bool operator >  (const lx_interval &, const real &) throw();
//! Implementation of standard greater-or-equal-than operation
inline bool operator >= (const lx_interval &, const real &) throw();

// ---- lx_interval--l_real

//! Implementation of standard less-than operation
inline bool operator <  (const l_real &, const lx_interval &) throw();
//! Implementation of standard less-or-equal-than operation
inline bool operator <= (const l_real &, const lx_interval &) throw();
//! Implementation of standard greater-than operation
inline bool operator >  (const lx_interval &, const l_real &) throw();
//! Implementation of standard greater-or-equal-than operation
inline bool operator >= (const lx_interval &, const l_real &) throw();

// ---- lx_interval--lx_real

//! Implementation of standard less-than operation
inline bool operator <  (const lx_real &, const lx_interval &) throw();
//! Implementation of standard less-or-equal-than operation
inline bool operator <= (const lx_real &, const lx_interval &) throw();
//! Implementation of standard greater-than operation
inline bool operator >  (const lx_interval &, const lx_real &) throw();
//! Implementation of standard greater-or-equal-than operation
inline bool operator >= (const lx_interval &, const lx_real &) throw();


// -------------------------- comparisons --------------------------------

//! Implementation of standard negation operation
inline bool operator ! (const lx_interval &) throw();

//! Implementation of standard equality operation
inline bool operator == (const lx_interval &, const lx_interval &) throw();
//! Implementation of standard equality operation
inline bool operator == (const lx_interval &, const l_interval &) throw();
//! Implementation of standard equality operation
inline bool operator == (const l_interval &, const lx_interval &) throw();
//! Implementation of standard equality operation
inline bool operator == (const lx_interval &, const interval &) throw();
//! Implementation of standard equality operation
inline bool operator == (const interval &, const lx_interval &) throw();
//! Implementation of standard equality operation
inline bool operator == (const lx_interval &, const real &) throw();
//! Implementation of standard equality operation
inline bool operator == (const real &, const lx_interval &) throw();
//! Implementation of standard equality operation
inline bool operator == (const lx_interval &, const l_real &) throw();
//! Implementation of standard equality operation
inline bool operator == (const l_real &, const lx_interval &) throw();
//! Implementation of standard equality operation
inline bool operator == (const lx_interval &, const lx_real &) throw();
//! Implementation of standard equality operation
inline bool operator == (const lx_real &, const lx_interval &) throw();

//! Implementation of standard negated equality operation
inline bool operator != (const lx_interval &, const lx_interval &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const lx_interval &, const l_interval &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const l_interval &, const lx_interval &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const lx_interval &, const interval &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const interval &, const lx_interval &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const lx_interval &, const real &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const real &, const lx_interval &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const lx_interval &, const l_real &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const l_real &, const lx_interval &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const lx_interval &, const lx_real &) throw();
//! Implementation of standard negated equality operation
inline bool operator != (const lx_real &, const lx_interval &) throw();

//! Implementation of standard algebraic positive sign operation
inline lx_interval operator+(const lx_interval & a) throw();
//! Implementation of standard algebraic negative sign operation
inline lx_interval operator-(const lx_interval & a) throw();

lx_interval operator + (const lx_interval &, const lx_interval &) throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const lx_interval &, const l_interval &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const l_interval &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const lx_interval &, const l_real &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const l_real &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const lx_interval &, const lx_real &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const lx_real &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const lx_interval &, const real &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const real &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const lx_interval &, const interval &) 
                                                                 throw();
//! Implementation of standard algebraic addition operation
inline lx_interval operator + (const interval &, const lx_interval &)
                                                                 throw();
//! Implementation of standard algebraic addition and allocation operation
inline lx_interval & operator +=(lx_interval &, const lx_interval &) throw();
//! Implementation of standard algebraic addition and allocation operation
inline lx_interval & operator +=(lx_interval &, const l_interval &) throw();
//! Implementation of standard algebraic addition and allocation operation
inline lx_interval & operator +=(lx_interval &, const l_real     &) throw();
//! Implementation of standard algebraic addition and allocation operation
inline lx_interval & operator +=(lx_interval &, const lx_real     &) throw();
//! Implementation of standard algebraic addition and allocation operation
inline lx_interval & operator +=(lx_interval &, const real       &) throw();
//! Implementation of standard algebraic addition and allocation operation
inline lx_interval & operator +=(lx_interval &, const interval   &) throw();

//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const lx_interval &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const lx_interval &, const l_interval &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const l_interval &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const lx_interval &, const l_real &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const l_real &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const lx_interval &, const lx_real &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const lx_real &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const lx_interval &, const real &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const real &, const lx_interval &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const lx_interval &, const interval &) 
                                                                 throw();
//! Implementation of standard algebraic subtraction operation
inline lx_interval operator - (const interval &, const lx_interval &) 
                                                                 throw();

//! Implementation of standard algebraic subtraction and allocation operation
inline lx_interval & operator -=(lx_interval &, const lx_interval &) throw();
//! Implementation of standard algebraic subtraction and allocation operation
inline lx_interval & operator -=(lx_interval &, const l_interval &) throw();
//! Implementation of standard algebraic subtraction and allocation operation
inline lx_interval & operator -=(lx_interval &, const l_real     &) throw();
//! Implementation of standard algebraic subtraction and allocation operation
inline lx_interval & operator -=(lx_interval &, const lx_real     &) throw();
//! Implementation of standard algebraic subtraction and allocation operation
inline lx_interval & operator -=(lx_interval &, const real       &) throw();
//! Implementation of standard algebraic subtraction and allocation operation
inline lx_interval & operator -=(lx_interval &, const interval   &) throw();

//! Implementation of standard algebraic multiplication operation
lx_interval operator * (const lx_interval &, const lx_interval &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const lx_interval &, const l_interval &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const l_interval &, const lx_interval &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const lx_interval &, const l_real &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const l_real &, const lx_interval &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const lx_interval &, const lx_real &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const lx_real &, const lx_interval &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const lx_interval &, const real &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const real &, const lx_interval &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const lx_interval &, const interval &) 
    throw();
//! Implementation of standard algebraic multiplication operation
inline lx_interval operator * (const interval &, const lx_interval &) 
    throw();

//! Implementation of standard algebraic multiplication and allocation operation
inline lx_interval & operator *=(lx_interval &, const lx_interval &) throw();
//! Implementation of standard algebraic multiplication and allocation operation
inline lx_interval & operator *=(lx_interval &, const l_interval &) throw();
//! Implementation of standard algebraic multiplication and allocation operation
inline lx_interval & operator *=(lx_interval &, const l_real     &) throw();
//! Implementation of standard algebraic multiplication and allocation operation
inline lx_interval & operator *=(lx_interval &, const lx_real     &) throw();
//! Implementation of standard algebraic multiplication and allocation operation
inline lx_interval & operator *=(lx_interval &, const real       &) throw();
//! Implementation of standard algebraic multiplication and allocation operation
inline lx_interval & operator *=(lx_interval &, const interval   &) throw();

//! Implementation of standard algebraic division operation
lx_interval operator / (const lx_interval &, const lx_interval &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const lx_interval &, const l_interval &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const l_interval &, const lx_interval &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const lx_interval &, const l_real &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const l_real &, const lx_interval &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const lx_interval &, const real &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const real &, const lx_interval &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const lx_interval &, const interval &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const interval &, const lx_interval &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const lx_interval &, const lx_real &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);
//! Implementation of standard algebraic division operation
inline lx_interval operator / (const lx_real &, const lx_interval &) 
    throw(ERROR_LINTERVAL_DIV_BY_ZERO);

//! Implementation of standard algebraic division and allocation operation
inline lx_interval & operator /=(lx_interval &, const lx_interval &) throw();
//! Implementation of standard algebraic division and allocation operation
inline lx_interval & operator /=(lx_interval &, const l_interval &) throw();
//! Implementation of standard algebraic division and allocation operation
inline lx_interval & operator /=(lx_interval &, const l_real     &) throw();
//! Implementation of standard algebraic division and allocation operation
inline lx_interval & operator /=(lx_interval &, const lx_real     &) throw();
//! Implementation of standard algebraic division and allocation operation
inline lx_interval & operator /=(lx_interval &, const real       &) throw();
//! Implementation of standard algebraic division and allocation operation
inline lx_interval & operator /=(lx_interval &, const interval   &) throw();

// ----------------------------- Convex hull -------------------------------

    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const lx_interval&, const lx_interval&) 
	throw();
    //! Returns the convex hull of the arguments 
    inline lx_interval operator | (const lx_interval&, const l_interval&) 
	throw();
    //! Returns the convex hull of the arguments 
    inline lx_interval operator | (const l_interval&, const lx_interval&) 
	throw();
    //! Returns the convex hull of the arguments 
    inline lx_interval operator | (const lx_interval&, const interval&) 
	throw();
    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const interval&, const lx_interval&) 
	throw();
    //! Allocates the convex hull of the arguments to the first argument
    inline lx_interval & operator |= (lx_interval&, const lx_interval&) 
	throw();
    //! Allocates the convex hull of the arguments to the first argument
    inline lx_interval & operator |= (lx_interval&, const l_interval&) 
	throw();
    //! Allocates the convex hull of the arguments to the first argument
    inline lx_interval & operator |= (lx_interval&, const interval&) 
	throw();
    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const lx_real&, const lx_interval&) 
	throw();
    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const real&, const lx_interval&) 
	throw();
    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const lx_interval&, const lx_real&) 
	throw();
    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const lx_interval&, const real&) 
	throw();
    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const lx_interval&, const l_real&) 
	throw();
    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const l_real&, const lx_interval&) 
	throw();
    //! Allocates the convex hull of the arguments to the first argument 
    inline lx_interval & operator |= (lx_interval&, const real&)
	throw();
    //! Allocates the convex hull of the arguments to the first argument
    inline lx_interval & operator |= (lx_interval&, const l_real&)
	throw();
    //! Allocates the convex hull of the arguments to the first argument
    inline lx_interval & operator |= (lx_interval&, const lx_real&)
	throw();
    //! Returns the convex hull of the arguments
    inline lx_interval operator | (const lx_real&, const lx_real&) 
	throw();

// --------------------------- Intersection -----------------------------

    //! Returns the intersection of the arguments
    inline lx_interval operator & (const lx_interval&, const lx_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
    //! Returns the intersection of the arguments
    inline lx_interval operator & (const lx_interval&, const l_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 
    //! Allocates the intersection of the arguments to the first argument
    inline lx_interval & operator &= (lx_interval&, const l_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 
    //! Returns the intersection of the arguments
    inline lx_interval operator & (const l_interval&, const lx_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
    //! Returns the intersection of the arguments 
    inline lx_interval operator & (const lx_interval&, const interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
    //! Allocates the intersection of the arguments to the first argument 
    inline lx_interval & operator &= (lx_interval &a, const interval &b) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 
    //! Returns the intersection of the arguments
    inline lx_interval operator & (const interval&, const lx_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
    //! Allocates the intersection of the arguments to the first argument 
    inline lx_interval & operator &= (lx_interval&, const lx_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 
    //! Returns the intersection of the arguments
    inline lx_interval operator & (const lx_interval&, const lx_real&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
    //! Returns the intersection of the arguments
    inline lx_interval operator & (const lx_interval&, const l_real&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 
    //! Returns the intersection of the arguments
    inline lx_interval operator & (const lx_interval&, const real&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
    //! Returns the intersection of the arguments 
    inline lx_interval operator & (const lx_real&, const lx_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
    //! Returns the intersection of the arguments
    inline lx_interval operator & (const l_real&, const lx_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 
    //! Returns the intersection of the arguments
    inline lx_interval operator & (const real&, const lx_interval&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
    //! Allocates the intersection of the arguments to the first argument 
    inline lx_interval & operator &= (lx_interval&, const lx_real&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 
    //! Allocates the intersection of the arguments to the first argument
    inline lx_interval & operator &= (lx_interval&, const l_real&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 
    //! Allocates the intersection of the arguments to the first argument
    inline lx_interval & operator &= (lx_interval&, const real&) 
	throw(ERROR_LINTERVAL_EMPTY_INTERVAL); 

// ------------------------- SetInf, SetSup -----------------------------

//! Returns the interval with the new given infimum value
inline lx_interval & SetInf(lx_interval&, const lx_real&) 
    throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
//! Returns the interval with the new given infimum value
inline lx_interval & SetInf(lx_interval&, const l_real&) 
    throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
//! Returns the interval with the new given infimum value
inline lx_interval & SetInf(lx_interval&, const real&) 
    throw(ERROR_LINTERVAL_EMPTY_INTERVAL);

//! Returns the interval with the new given supremum value
inline lx_interval & SetSup(lx_interval&, const lx_real&) 
    throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
//! Returns the interval with the new given supremum value
inline lx_interval & SetSup(lx_interval&, const l_real&) 
    throw(ERROR_LINTERVAL_EMPTY_INTERVAL);
//! Returns the interval with the new given supremum value
inline lx_interval & SetSup(lx_interval&, const real&) 
    throw(ERROR_LINTERVAL_EMPTY_INTERVAL);

// ------------------------- Help Functions: ----------------------------

/* ---------------- Intervals for test calculations ------------------ */

//! Returns a point interval with maximum exponent 1020
l_interval point_max(void);
// returns a staggered point interval with maximum exponent 1020,
// whereby nearly all mantissa bits are set. 

//! Returns a point interval with exponent n
l_interval point_any(int n);
// returns a staggered point interval with exponent n,
// whereby nearly all mantissa bits are set.
// -1074 <= n <= +1020; 

//! Returns a wide interval with maximum exponent 1020
l_interval wide_max(void);
// returns a staggered interval a with maximum exponent 1020
// and diam(a)>0, whereby nearly all mantissa bits are set. 

//! Returns a wide interval with exponent n, \f$ -1074\le n \le +1020; \f$
l_interval wide_any(int n);
// returns a wide staggered interval a with exponent n,
// and diam(a)>0, whereby nearly all mantissa bits are set.
// -1074 <= n <= +1020;

// void Dec2Bin(int p, const l_interval &x, int &ex, l_interval &li) throw();

/* ---------------- Array of constants ------------------- */

extern real ln_N[180];
// ln_N[0] = ln(2); ln_N[1] = ln(3); ... ln_N[179] = ln(181); 
// Implemented in lx_interval.cpp;

/* ------------------------------------------------------- */

}  // end namespace cxsc

#include "lx_interval.inl"

#endif
