/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.2)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  F. Blomquist, University of Wuppertal, 20.09.2007;
*/

/* CVS $Id: lx_real.cpp,v 1.14 2006/12/06 19:14:09 cxsc Exp $ */

#include <memory.h>
#include "lx_real.hpp"

namespace cxsc {

bool operator == (const lx_real &a, const lx_real &b) throw()
{
    l_real ar(lr_part(a)), br(lr_part(b));
    int na(expo(a)), nb(expo(b)), exa(expo_gr(ar)), exb(expo_gr(br)),d;
    bool a_0(exa<-100000), b_0(exb<-100000), res;
    interval z(0);

    if (a_0 || b_0) res = (a_0 == b_0);
    else // a,b <> 0:
    {
	d = exa-exb;
	if (d>0)
	{  // br nach oben skalieren: (Overflow dabei nicht moeglich!)
	    Times2pown(br,z,d);
	    nb = sub_int(nb,d);  // b = 2^nb * br;
	} else // d<=0:
	{  // ar nach oben skalieren: (Overflow dabei nicht moeglich!)
	    Times2pown(ar,z,-d);
	    na = add_int(na,d);
	}
	res = (na==nb && ar==br);
    }
    return res;
}

bool operator > (const lx_real &a, const lx_real &b) throw()
{
    bool zero_a(eq_zero(a)), zero_b(eq_zero(b)), bl(false);
    int na,nb,ex,d;
    l_real lr;
    interval z;
    if (zero_a) return sm_zero(b);
    if (zero_b) return gr_zero(a);
    // Now we have:  a,b != 0:
    na = expo(a);  nb = expo(b);
    if (na>nb)
    {
	if (na>0 && nb<0 && na>nb+5000)
	    bl = gr_zero(a);
	else 
	{
	    lr = lr_part(a);
	    d = sub_int(na,nb); // d>0;
	    ex = expo_gr(lr);
	    if (add_int(ex,d)<=1024) // Kein Overflow!
	    {
		Times2pown(lr,z,d);
		bl = lr > lr_part(b);
	    } else bl = lr>0;  // Overflow!
	}
    } else // nb >= na:
    {
	if (nb>0 && na<0 && nb>na+5000)
	    bl = sm_zero(b);
	else
	{
	    lr = lr_part(b);
	    d = sub_int(nb,na); // d>=0; // Hier kann int-Overflow entstehen!
	    ex = expo_gr(lr);
	    if (add_int(ex,d)<=1024) // Kein Overflow!
	    {
		if (d>0)
		    Times2pown(lr,z,d);
		bl = lr_part(a) > lr;
	    } else bl = lr<0;  // Overflow!
	}
    }

    return bl;
}

void scale_up(lx_real &a) throw()
// scale_up skaliert a.lr moeglichst weit und nur nach oben.  
// 2^(a.ex)*a.lr  hat fuer d>=0 vor und nach der Skalierung stets
// den gleichen Wert.  
{
    interval z(0);
    int d;
    d = 1023 - expo_gr(a.lr);  // d >= 0, wenn a.lr <> 0;
    if (d>0) // If (a==0) then d<0 and nothing is done!
    {
	Times2pown(a.lr,z,d);   // a.lr wird mit 2^d multipliziert.
	a.ex = sub_int(a.ex,d); // damit 2^(a.ex)*a.li konstant bleibt.
    }
}


} // end: namespace cxsc
